"""Detect component version change availability and broadcast on MQTT topic, suitable for HomeAssistant autodetect"""

from usingversion import getattr_with_version

__getattr__ = getattr_with_version("updates2mqtt", __file__, __name__)



import tempfile
import uuid
from pathlib import Path

import pytest
from omegaconf import OmegaConf

from updates2mqtt.config import MqttConfig, load_app_config, load_package_info

EXAMPLES_ROOT = "examples"
examples = [str(p.name) for p in Path(EXAMPLES_ROOT).iterdir() if p.name.startswith("config")]


def test_envvar_config(monkeypatch) -> None:  # noqa: ANN001
    with tempfile.TemporaryDirectory() as tmpdir:
        monkeypatch.setenv("MQTT_HOST", "193.11.55.12")
        monkeypatch.setenv("MQTT_USER", "tester")
        monkeypatch.setenv("MQTT_PASS", uuid.uuid4().hex)
        monkeypatch.setenv("MQTT_PORT", "1824")
        conf_path: Path = Path(tmpdir) / "config-test-env.yaml"
        validated_config = load_app_config(conf_path, return_invalid=True)
        assert validated_config is not None
        assert validated_config.node.git_path == "/usr/bin/git"
        assert validated_config.mqtt.port == 1824
        assert validated_config.mqtt.host == "193.11.55.12"
        assert validated_config.mqtt.user == "tester"


@pytest.mark.parametrize("config_name", examples)
def test_config(config_name: str, monkeypatch) -> None:  # noqa: ANN001
    monkeypatch.setenv("MQTT_USER", "tester")
    monkeypatch.setenv("MQTT_PASS", uuid.uuid4().hex)
    config_path: Path = Path(EXAMPLES_ROOT) / config_name
    validated_config = load_app_config(config_path)
    assert validated_config is not None
    assert validated_config.node.git_path == "/usr/bin/git"


def test_round_trip_config() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        conf_path: Path = Path(tmpdir) / "config-test.yaml"
        generated_config = load_app_config(conf_path, return_invalid=True)
        assert conf_path.exists()
        assert generated_config is not None
        # set mandatory values
        OmegaConf.set_readonly(generated_config, False)  # type: ignore[arg-type]
        generated_config.mqtt = MqttConfig(
            host=generated_config.mqtt.host,
            user="myuser",
            password=uuid.uuid4().hex,
            port=generated_config.mqtt.port,
            topic_root=generated_config.mqtt.topic_root,
        )
        conf_path.write_text(OmegaConf.to_yaml(generated_config))
        reloaded_config = load_app_config(conf_path)
        assert reloaded_config is not None
        assert reloaded_config.mqtt.user == "myuser"


def test_package_config() -> None:
    validated_pkg_info = load_package_info(Path("common_packages.yaml"))
    assert validated_pkg_info is not None
    assert len(validated_pkg_info.common_packages) > 0



import os
import typing
from dataclasses import dataclass, field
from pathlib import Path

import structlog
from omegaconf import MISSING, DictConfig, MissingMandatoryValue, OmegaConf, ValidationError

log = structlog.get_logger()


@dataclass
class MqttConfig:
    host: str = "${oc.env:MQTT_HOST,localhost}"
    user: str = f"${{oc.env:MQTT_USER,{MISSING}}}"
    password: str = f"${{oc.env:MQTT_PASS,{MISSING}}}"
    port: int = "${oc.decode:${oc.env:MQTT_PORT,1883}}"  # type: ignore[assignment]
    topic_root: str = "updates2mqtt"
    protocol: str = "3.11"


@dataclass
class MetadataSourceConfig:
    enabled: bool = True
    cache_ttl: int = 60 * 60 * 24 * 7  # 1 week


@dataclass
class DockerConfig:
    enabled: bool = True
    allow_pull: bool = True
    allow_restart: bool = True
    allow_build: bool = True
    compose_version: str = "v2"
    default_entity_picture_url: str = "https://www.docker.com/wp-content/uploads/2022/03/Moby-logo.png"
    device_icon: str = "mdi:docker"  # Icon to show when browsing entities in Home Assistant
    discover_metadata: dict[str, MetadataSourceConfig] = field(
        default_factory=lambda: {"linuxserver.io": MetadataSourceConfig(enabled=True)}
    )


@dataclass
class HomeAssistantDiscoveryConfig:
    prefix: str = "homeassistant"
    enabled: bool = True


@dataclass
class HomeAssistantConfig:
    discovery: HomeAssistantDiscoveryConfig = field(default_factory=HomeAssistantDiscoveryConfig)
    state_topic_suffix: str = "state"


@dataclass
class HealthCheckConfig:
    enabled: bool = True
    interval: int = 300  # Interval in seconds to publish healthcheck message, 0 to disable
    topic_template: str = "healthcheck/{node_name}/updates2mqtt"


@dataclass
class NodeConfig:
    name: str = field(default_factory=lambda: os.uname().nodename.replace(".local", ""))
    git_path: str = "/usr/bin/git"
    healthcheck: HealthCheckConfig = field(default_factory=HealthCheckConfig)


@dataclass
class LogConfig:
    level: str = "INFO"


@dataclass
class Config:
    log: LogConfig = field(default_factory=LogConfig)
    node: NodeConfig = field(default_factory=NodeConfig)
    mqtt: MqttConfig = field(default_factory=MqttConfig)
    homeassistant: HomeAssistantConfig = field(default_factory=HomeAssistantConfig)
    docker: DockerConfig = field(default_factory=DockerConfig)
    scan_interval: int = 60 * 60 * 3


@dataclass
class DockerPackageUpdateInfo:
    image_name: str = MISSING


@dataclass
class PackageUpdateInfo:
    docker: DockerPackageUpdateInfo | None = field(default_factory=DockerPackageUpdateInfo)
    logo_url: str | None = None
    release_notes_url: str | None = None


@dataclass
class UpdateInfoConfig:
    common_packages: dict[str, PackageUpdateInfo] = field(default_factory=lambda: {})


class IncompleteConfigException(BaseException):
    pass


def load_package_info(pkginfo_file_path: Path) -> UpdateInfoConfig:
    if pkginfo_file_path.exists():
        log.debug("Loading common package update info", path=pkginfo_file_path)
        cfg = OmegaConf.load(pkginfo_file_path)
    else:
        log.warn("No common package update info found", path=pkginfo_file_path)
        cfg = OmegaConf.structured(UpdateInfoConfig)
    OmegaConf.set_readonly(cfg, True)
    return typing.cast("UpdateInfoConfig", cfg)


def load_app_config(conf_file_path: Path, return_invalid: bool = False) -> Config | None:
    base_cfg: DictConfig = OmegaConf.structured(Config)
    if conf_file_path.exists():
        cfg: DictConfig = typing.cast("DictConfig", OmegaConf.merge(base_cfg, OmegaConf.load(conf_file_path)))
    else:
        if not conf_file_path.parent.exists():
            try:
                log.debug(f"Creating config directory {conf_file_path.parent} if not already present")
                conf_file_path.parent.mkdir(parents=True, exist_ok=True)
            except Exception:
                log.exception("Unable to create config directory", path=conf_file_path.parent)
        try:
            conf_file_path.write_text(OmegaConf.to_yaml(base_cfg))
            log.info(f"Auto-generated a new config file at {conf_file_path}")
        except Exception:
            log.exception("Unable to write config file", path=conf_file_path)
        cfg = base_cfg

    try:
        # Validate that all required fields are present, throw exception now rather than when config first used
        OmegaConf.to_container(cfg, throw_on_missing=True)
        OmegaConf.set_readonly(cfg, True)
        config: Config = typing.cast("Config", cfg)

        if config.mqtt.user in ("", MISSING) or config.mqtt.password in ("", MISSING):
            log.info("The config has place holders for MQTT user and/or password")
            if not return_invalid:
                return None
        return config
    except (MissingMandatoryValue, ValidationError) as e:
        log.error("Configuration error %s", e, path=conf_file_path.as_posix())
        if return_invalid and cfg is not None:
            return typing.cast("Config", cfg)
        raise



import asyncio
import logging
import sys
import time
import uuid
from collections.abc import Callable
from datetime import UTC, datetime
from pathlib import Path
from threading import Event
from typing import Any

import structlog

import updates2mqtt
from updates2mqtt.model import Discovery, ReleaseProvider

from .config import Config, load_app_config, load_package_info
from .integrations.docker import DockerProvider
from .mqtt import MqttPublisher

log = structlog.get_logger()

CONF_FILE = Path("conf/config.yaml")
PKG_INFO_FILE = Path("./common_packages.yaml")
UPDATE_INTERVAL = 60 * 60 * 4

# #TODO:
#  - Set install in progress
#  - Support apt
#  - Retry on registry fetch fail
#  - Fetcher in subproc or thread
#  - Clear command message after install
#  - use git hash as alt to img ref for builds, or daily builds


class App:
    def __init__(self) -> None:
        self.startup_timestamp: str = datetime.now(UTC).isoformat()
        self.last_scan_timestamp: str | None = None
        app_config: Config | None = load_app_config(CONF_FILE)
        if app_config is None:
            log.error(f"Invalid configuration at {CONF_FILE}, edit config to fix missing or invalid values and restart")
            log.error("Exiting app")
            sys.exit(1)
        self.cfg: Config = app_config

        structlog.configure(wrapper_class=structlog.make_filtering_bound_logger(getattr(logging, self.cfg.log.level)))
        log.debug("Logging initialized", level=self.cfg.log.level)
        self.common_pkg = load_package_info(PKG_INFO_FILE)

        self.publisher = MqttPublisher(self.cfg.mqtt, self.cfg.node, self.cfg.homeassistant)

        self.scanners: list[ReleaseProvider] = []
        self.scan_count: int = 0
        self.last_scan: str | None = None
        if self.cfg.docker.enabled:
            self.scanners.append(DockerProvider(self.cfg.docker, self.common_pkg, self.cfg.node))
        self.stopped = Event()
        self.healthcheck_topic = self.cfg.node.healthcheck.topic_template.format(node_name=self.cfg.node.name)

        log.info(
            "App configured",
            node=self.cfg.node.name,
            scan_interval=self.cfg.scan_interval,
        )

    async def scan(self) -> None:
        session = uuid.uuid4().hex
        for scanner in self.scanners:
            log.info("Cleaning topics before scan", source_type=scanner.source_type)
            if self.scan_count == 0:
                await self.publisher.clean_topics(scanner, None, force=True)
            if self.stopped.is_set():
                break
            log.info("Scanning", source=scanner.source_type, session=session)
            async with asyncio.TaskGroup() as tg:
                async for discovery in scanner.scan(session):  # xtype: ignore[attr-defined]
                    tg.create_task(self.on_discovery(discovery), name=f"discovery-{discovery.name}")
            if self.stopped.is_set():
                break
            await self.publisher.clean_topics(scanner, session, force=False)
            self.scan_count += 1
            log.info("Scan complete", source_type=scanner.source_type)
        self.last_scan_timestamp = datetime.now(UTC).isoformat()

    async def main_loop(self) -> None:
        log.debug("Starting run loop")
        self.publisher.start()

        if self.cfg.node.healthcheck.enabled:
            await self.healthcheck()  # initial eager healthcheck
            log.info(
                f"Setting up healthcheck every {self.cfg.node.healthcheck.interval} seconds to topic {self.healthcheck_topic}"
            )
            self.healthcheck_loop_task = asyncio.create_task(
                repeated_call(self.healthcheck, interval=self.cfg.node.healthcheck.interval), name="healthcheck"
            )

        for scanner in self.scanners:
            self.publisher.subscribe_hass_command(scanner)

        while not self.stopped.is_set() and self.publisher.is_available():
            await self.scan()
            if not self.stopped.is_set() and self.publisher.is_available():
                await asyncio.sleep(self.cfg.scan_interval)
            else:
                log.info("Stop requested, exiting run loop and skipping sleep")

        if not self.publisher.is_available():
            log.error("MQTT fatal connection error - check host,port,user,password in config")
            self.shutdown(exit_code=1)

        log.debug("Exiting run loop")

    async def on_discovery(self, discovery: Discovery) -> None:
        dlog = log.bind(name=discovery.name)
        try:
            if self.cfg.homeassistant.discovery.enabled:
                self.publisher.publish_hass_config(discovery)

            self.publisher.publish_hass_state(discovery)
            if discovery.update_policy == "Auto":
                # TODO: review auto update, trigger by version, use update interval as throttle
                elapsed: float = (
                    time.time() - discovery.update_last_attempt if discovery.update_last_attempt is not None else -1
                )
                if elapsed == -1 or elapsed > UPDATE_INTERVAL:
                    dlog.info(
                        "Initiate auto update (last:%s, elapsed:%s, max:%s)",
                        discovery.update_last_attempt,
                        elapsed,
                        UPDATE_INTERVAL,
                    )
                    self.publisher.local_message(discovery, "install")
                else:
                    dlog.info("Skipping auto update")
        except asyncio.CancelledError:
            dlog.info("Discovery handling cancelled")
        except Exception:
            dlog.exception("Discovery handling failed")
            raise

    async def interrupt_tasks(self) -> None:
        running_tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        log.info(f"Cancelling {len(running_tasks)} tasks")
        for t in running_tasks:
            log.debug("Cancelling task", task=t.get_name())
            if t.get_name() == "healthcheck" or t.get_name().startswith("discovery-"):
                t.cancel()
        await asyncio.gather(*running_tasks, return_exceptions=True)
        log.debug("Cancellation task completed")

    def shutdown(self, *args, exit_code: int = 143) -> None:  # noqa: ANN002, ARG002
        log.info("Shutting down, exit_code: %s", exit_code)
        self.stopped.set()
        for scanner in self.scanners:
            scanner.stop()
        interrupt_task = asyncio.get_event_loop().create_task(
            self.interrupt_tasks(),
            eager_start=True,  # type: ignore[call-arg] # pyright: ignore[reportCallIssue]
            name="interrupt",
        )
        for t in asyncio.all_tasks():
            log.debug("Tasks waiting = %s", t)
        self.publisher.stop()
        log.debug("Interrupt: %s", interrupt_task.done())
        log.info("Shutdown handling complete")
        sys.exit(exit_code)  # SIGTERM Graceful Exit = 143

    async def healthcheck(self) -> None:
        if not self.publisher.is_available():
            return
        self.publisher.publish(
            topic=self.healthcheck_topic,
            payload={
                "version": updates2mqtt.version,  # pyright: ignore[reportAttributeAccessIssue]
                "node": self.cfg.node.name,
                "heartbeat_raw": time.time(),
                "heartbeat_stamp": datetime.now(UTC).isoformat(),
                "startup_stamp": self.startup_timestamp,
                "last_scan_stamp": self.last_scan_timestamp,
                "scan_count": self.scan_count,
            },
        )


async def repeated_call(func: Callable, interval: int = 60, *args: Any, **kwargs: Any) -> None:
    # run a task periodically indefinitely
    while True:
        try:
            await func(*args, **kwargs)
            await asyncio.sleep(interval)
        except asyncio.CancelledError:
            log.debug("Periodic task cancelled")
        except Exception:
            log.exception("Periodic task failed")


def run() -> None:
    import asyncio
    import signal

    from .app import App

    log.debug(f"Starting updates2mqtt v{updates2mqtt.version}")  # pyright: ignore[reportAttributeAccessIssue]
    app = App()

    signal.signal(signal.SIGTERM, app.shutdown)
    try:
        asyncio.run(app.main_loop(), debug=False)
        log.debug("App exited gracefully")
    except asyncio.CancelledError:
        log.debug("App exited on cancelled task")


if __name__ == "__main__":
    run()



from typing import Any

import structlog

import updates2mqtt
from updates2mqtt.model import Discovery

log = structlog.get_logger()
HASS_UPDATE_SCHEMA = [
    "installed_version",
    "latest_version",
    "title",
    "release_summary",
    "release_url",
    "entity_picture",
    "in_progress",
    "update_percentage",
]


def hass_format_config(
    discovery: Discovery, object_id: str, node_name: str, state_topic: str, command_topic: str | None, session: str
) -> dict[str, Any]:
    config = {
        "name": f"{discovery.name} {discovery.source_type} on {node_name}",
        "device_class": None,  # not firmware, so defaults to null
        "unique_id": object_id,
        "state_topic": state_topic,
        "source_session": session,
        "supported_features": discovery.features,
        "entity_picture": discovery.entity_picture_url,
        "icon": discovery.device_icon,
        "can_update": discovery.can_update,
        "can_build": discovery.can_build,
        "can_restart": discovery.can_restart,
        "update_policy": discovery.update_policy,
        "latest_version_topic": state_topic,
        "latest_version_template": "{{value_json.latest_version}}",
        "origin": {
            "name": "updates2mqtt",
            "sw_version": updates2mqtt.version,
            "support_url": "https://github.com/rhizomatics/updates2mqtt/issues",
        },
    }
    if command_topic:
        config["command_topic"] = command_topic
        config["payload_install"] = f"{discovery.source_type}|{discovery.name}|install"
    if discovery.custom.get("git_repo_path"):
        config["git_repo_path"] = discovery.custom["git_repo_path"]
    config.update(discovery.provider.hass_config_format(discovery))
    return config


def hass_format_state(discovery: Discovery, node_name: str, session: str, in_progress: bool = False) -> dict[str, Any]:  # noqa: ARG001
    title: str = (
        discovery.title_template.format(name=discovery.name, node=node_name) if discovery.title_template else discovery.name
    )
    state = {
        "installed_version": discovery.current_version,
        "latest_version": discovery.latest_version,
        "title": title,
        "in_progress": in_progress,
    }
    if discovery.release_summary:
        state["release_summary"] = discovery.release_summary
    if discovery.release_url:
        state["release_url"] = discovery.release_url
    custom_state = discovery.provider.hass_state_format(discovery)
    if custom_state:
        state.update(custom_state)
    invalid_keys = [k for k in state if k not in HASS_UPDATE_SCHEMA]
    if invalid_keys:
        log.warning(f"Invalid keys in state: {invalid_keys}")
        state = {k: v for k, v in state.items() if k in HASS_UPDATE_SCHEMA}
    return state



import datetime
import subprocess
import time
import typing
from collections.abc import AsyncGenerator, Callable
from enum import Enum
from pathlib import Path
from typing import Any, cast

import docker
import docker.errors
import structlog
from docker.models.containers import Container
from hishel.httpx import SyncCacheClient

from updates2mqtt.config import DockerConfig, DockerPackageUpdateInfo, NodeConfig, PackageUpdateInfo, UpdateInfoConfig
from updates2mqtt.model import Discovery, ReleaseProvider

from .git_utils import git_check_update_available, git_pull, git_timestamp, git_trust

if typing.TYPE_CHECKING:
    from docker.models.images import Image, RegistryData

# distinguish docker build from docker pull?

log = structlog.get_logger()
NO_KNOWN_IMAGE = "UNKNOWN"


class DockerComposeCommand(Enum):
    BUILD = "build"
    UP = "up"


def safe_json_dt(t: float | None) -> str | None:
    return time.strftime("%Y-%m-%dT%H:%M:%S.0000", time.gmtime(t)) if t else None


class DockerProvider(ReleaseProvider):
    def __init__(self, cfg: DockerConfig, common_pkg_cfg: UpdateInfoConfig, node_cfg: NodeConfig) -> None:
        super().__init__("docker")
        self.client: docker.DockerClient = docker.from_env()
        self.cfg: DockerConfig = cfg
        self.node_cfg: NodeConfig = node_cfg
        self.common_pkgs: dict[str, PackageUpdateInfo] = common_pkg_cfg.common_packages if common_pkg_cfg else {}
        # TODO: refresh discovered packages periodically
        self.discovered_pkgs: dict[str, PackageUpdateInfo] = self.discover_metadata()

    def update(self, discovery: Discovery) -> bool:
        logger: Any = self.log.bind(container=discovery.name, action="update")
        logger.info("Updating - last at %s", discovery.update_last_attempt)
        discovery.update_last_attempt = time.time()
        self.fetch(discovery)
        restarted = self.restart(discovery)
        logger.info("Updated - recorded at %s", discovery.update_last_attempt)
        return restarted

    def fetch(self, discovery: Discovery) -> None:
        logger = self.log.bind(container=discovery.name, action="fetch")

        image_ref: str | None = discovery.custom.get("image_ref")
        platform: str | None = discovery.custom.get("platform")
        if discovery.custom.get("can_pull") and image_ref:
            logger.info("Pulling", image_ref=image_ref, platform=platform)
            image: Image = self.client.images.pull(image_ref, platform=platform, all_tags=False)
            if image:
                logger.info("Pulled", image_id=image.id, image_ref=image_ref, platform=platform)
            else:
                logger.warn("Unable to pull", image_ref=image_ref, platform=platform)
        elif discovery.can_build:
            compose_path: str | None = discovery.custom.get("compose_path")
            git_repo_path: str | None = discovery.custom.get("git_repo_path")
            if not compose_path or not git_repo_path:
                logger.warn("No compose path or git repo path configured, skipped build")
                return
            if compose_path and not Path(git_repo_path).is_absolute():
                full_repo_path: Path = Path(compose_path) / git_repo_path
            else:
                full_repo_path = Path(git_repo_path)
            if git_check_update_available(full_repo_path, Path(self.node_cfg.git_path)):
                git_pull(full_repo_path, Path(self.node_cfg.git_path))
            if compose_path:
                self.build(discovery, compose_path)
            else:
                logger.warn("No compose path configured, skipped build")

    def build(self, discovery: Discovery, compose_path: str) -> bool:
        logger = self.log.bind(container=discovery.name, action="build")
        logger.info("Building")
        return self.execute_compose(DockerComposeCommand.BUILD, "", compose_path, logger)

    def execute_compose(self, command: DockerComposeCommand, args: str, cwd: str | None, logger: structlog.BoundLogger) -> bool:
        if not cwd or not Path(cwd).is_dir():
            logger.warn("Invalid compose path, skipped %s", command)
            return False
        logger.info(f"Executing compose {command} {args}")
        cmd: str = "docker-compose" if self.cfg.compose_version == "v1" else "docker compose"
        cmd = cmd + " " + command.value
        if args:
            cmd = cmd + " " + args

        proc = subprocess.run(cmd, check=False, shell=True, cwd=cwd)
        if proc.returncode == 0:
            logger.info(f"{command} via compose successful")
            return True
        logger.warn(
            f"{command} failed: %s",
            proc.returncode,
        )
        return False

    def restart(self, discovery: Discovery) -> bool:
        logger = self.log.bind(container=discovery.name, action="restart")
        compose_path = discovery.custom.get("compose_path")
        return self.execute_compose(DockerComposeCommand.UP, "--detach --yes", compose_path, logger)

    def rescan(self, discovery: Discovery) -> Discovery | None:
        logger = self.log.bind(container=discovery.name, action="rescan")
        try:
            c: Container = self.client.containers.get(discovery.name)
            if c:
                rediscovery = self.analyze(c, discovery.session, original_discovery=discovery)
                if rediscovery:
                    self.discoveries[rediscovery.name] = rediscovery
                    return rediscovery
            logger.warn("Unable to find container for rescan")
        except docker.errors.NotFound:
            logger.warn("Container not found in Docker")
        except docker.errors.APIError:
            logger.exception("Docker API error retrieving container")
        return None

    def analyze(self, c: Container, session: str, original_discovery: Discovery | None = None) -> Discovery | None:
        logger = self.log.bind(container=c.name, action="analyze")
        image_ref = None
        image_name = None
        local_versions = None
        if c.attrs is None or not c.attrs:
            logger.warn("No container attributes found, discovery rejected")
            return None
        if c.name is None:
            logger.warn("No container name found, discovery rejected")
            return None

        def env_override(env_var: str, default: Any) -> Any | None:
            return default if c_env.get(env_var) is None else c_env.get(env_var)

        env_str = c.attrs.get("Config", {}).get("Env")
        c_env = dict(env.split("=", maxsplit=1) for env in env_str if "==" not in env)
        ignore_container: str | None = env_override("UPD2MQTT_IGNORE", "FALSE")
        if ignore_container and ignore_container.upper() in ("1", "TRUE"):
            logger.info("Container ignored due to UPD2MQTT_IGNORE setting")
            return None

        image: Image | None = c.image
        if image is not None and image.tags and len(image.tags) > 0:
            image_ref = image.tags[0]
        else:
            image_ref = c.attrs.get("Config", {}).get("Image")
        if image_ref is None:
            logger.warn("No image or image attributes found")
        else:
            try:
                image_name = image_ref.split(":")[0]
            except Exception as e:
                logger.warn("No tags found (%s) : %s", image, e)
            if image is not None and image.attrs is not None:
                try:
                    local_versions = [i.split("@")[1][7:19] for i in image.attrs["RepoDigests"]]
                except Exception as e:
                    logger.warn("Cannot determine local version: %s", e)
                    logger.warn("RepoDigests=%s", image.attrs.get("RepoDigests"))

        platform: str = "Unknown"
        pkg_info: PackageUpdateInfo = self.default_metadata(image_name)

        try:
            picture_url = env_override("UPD2MQTT_PICTURE", pkg_info.logo_url)
            relnotes_url = env_override("UPD2MQTT_RELNOTES", pkg_info.release_notes_url)
            if image is not None and image.attrs is not None:
                platform = "/".join(
                    filter(
                        None,
                        [
                            image.attrs["Os"],
                            image.attrs["Architecture"],
                            image.attrs.get("Variant"),
                        ],
                    ),
                )

            reg_data: RegistryData | None = None
            latest_version: str | None = NO_KNOWN_IMAGE
            local_version: str | None = NO_KNOWN_IMAGE

            if image_ref and local_versions:
                retries_left = 3
                while reg_data is None and retries_left > 0 and not self.stopped.is_set():
                    try:
                        reg_data = self.client.images.get_registry_data(image_ref)
                        latest_version = reg_data.short_id[7:] if reg_data else None
                    except docker.errors.APIError as e:
                        retries_left -= 1
                        if retries_left == 0 or e.is_client_error():
                            logger.warn("Failed to fetch registry data: [%s] %s", e.errno, e.explanation)
                        else:
                            logger.debug("Failed to fetch registry data, retrying: %s", e)

            if local_versions:
                # might be multiple RepoDigests if image has been pulled multiple times with diff manifests
                local_version = latest_version if latest_version in local_versions else local_versions[0]

            def save_if_set(key: str, val: datetime.datetime | str | None) -> None:
                if val is not None:
                    custom[key] = val

            image_ref = image_ref or ""

            custom: dict[str, str | datetime.datetime | bool] = {}
            custom["platform"] = platform
            custom["image_ref"] = image_ref
            save_if_set("compose_path", c.labels.get("com.docker.compose.project.working_dir"))
            save_if_set("compose_version", c.labels.get("com.docker.compose.version"))
            save_if_set("git_repo_path", c_env.get("UPD2MQTT_GIT_REPO_PATH"))
            save_if_set("apt_pkgs", c_env.get("UPD2MQTT_APT_PKGS"))

            if c_env.get("UPD2MQTT_UPDATE") == "AUTO":
                logger.debug("Auto update policy detected")
                update_policy = "Auto"
            else:
                update_policy = "Passive"

            if custom.get("git_repo_path") and custom.get("compose_path"):
                full_repo_path: Path = Path(cast("str", custom.get("compose_path"))).joinpath(
                    cast("str", custom.get("git_repo_path"))
                )

                git_trust(full_repo_path, Path(self.node_cfg.git_path))
                save_if_set("git_local_timestamp", git_timestamp(full_repo_path, Path(self.node_cfg.git_path)))
            features: list[str] = []
            can_pull: bool = (
                self.cfg.allow_pull
                and image_ref is not None
                and image_ref != ""
                and (local_version != NO_KNOWN_IMAGE or latest_version != NO_KNOWN_IMAGE)
            )
            can_build: bool = self.cfg.allow_build and custom.get("git_repo_path") is not None
            can_restart: bool = self.cfg.allow_restart and custom.get("compose_path") is not None
            can_update: bool = False
            if self.cfg.allow_pull and not can_pull and not can_build:
                logger.info(
                    f"Pull not available, image_ref:{image_ref},local_version:{local_version},latest_version:{latest_version}"
                )
            if can_pull or can_build or can_restart:
                # public install-neutral capabilities and Home Assistant features
                can_update = True
                features.append("INSTALL")
                features.append("PROGRESS")
            elif any((self.cfg.allow_build, self.cfg.allow_restart, self.cfg.allow_pull)):
                logger.info(f"Update not available, can_pull:{can_pull}, can_build:{can_build},can_restart{can_restart}")
            if relnotes_url:
                features.append("RELEASE_NOTES")
            custom["can_pull"] = can_pull

            return Discovery(
                self,
                c.name,
                session,
                entity_picture_url=picture_url,
                release_url=relnotes_url,
                current_version=local_version,
                update_policy=update_policy,
                update_last_attempt=(original_discovery and original_discovery.update_last_attempt) or None,
                latest_version=latest_version if latest_version != NO_KNOWN_IMAGE else local_version,
                title_template="Docker image update for {name} on {node}",
                device_icon=self.cfg.device_icon,
                can_update=can_update,
                can_build=can_build,
                can_restart=can_restart,
                status=(c.status == "running" and "on") or "off",
                custom=custom,
                features=features,
            )
        except Exception:
            logger.exception("Docker Discovery Failure", container_attrs=c.attrs)
        return None

    async def scan(self, session: str) -> AsyncGenerator[Discovery]:
        logger = self.log.bind(session=session, action="scan")
        containers = results = 0
        for c in self.client.containers.list():
            if self.stopped.is_set():
                logger.info(f"Shutdown detected, aborting scan at {c}")
                break
            containers = containers + 1
            result = self.analyze(c, session)
            if result:
                self.discoveries[result.name] = result
                results = results + 1
                yield result
        logger.info("Completed", container_count=containers, result_count=results)

    def command(self, discovery_name: str, command: str, on_update_start: Callable, on_update_end: Callable) -> bool:
        logger = self.log.bind(container=discovery_name, action="command", command=command)
        logger.info("Executing")
        discovery: Discovery | None = None
        updated: bool = False
        try:
            discovery = self.resolve(discovery_name)
            if not discovery:
                logger.warn("Unknown entity", entity=discovery_name)
            elif command != "install":
                logger.warn("Unknown command")
            else:
                if discovery.can_update:
                    rediscovery: Discovery | None = None
                    logger.info("Starting update ...")
                    on_update_start(discovery)
                    if self.update(discovery):
                        logger.info("Rescanning ...")
                        rediscovery = self.rescan(discovery)
                        updated = rediscovery is not None
                        logger.info("Rescanned %s: %s", updated, rediscovery)
                    else:
                        logger.info("Rescan with no result")
                    on_update_end(rediscovery or discovery)
                else:
                    logger.warning("Update not supported for this container")
        except Exception:
            logger.exception("Failed to handle", discovery_name=discovery_name, command=command)
            if discovery:
                on_update_end(discovery)
        return updated

    def resolve(self, discovery_name: str) -> Discovery | None:
        return self.discoveries.get(discovery_name)

    def hass_state_format(self, discovery: Discovery) -> dict:  # noqa: ARG002
        # disable since hass mqtt update has strict json schema for message
        return {
            # "docker_image_ref": discovery.custom.get("image_ref"),
            # "last_update_attempt": safe_json_dt(discovery.update_last_attempt),
            # "can_pull": discovery.custom.get("can_pull"),
            # "can_build": discovery.custom.get("can_build"),
            # "can_restart": discovery.custom.get("can_restart"),
            # "git_repo_path": discovery.custom.get("git_repo_path"),
            # "compose_path": discovery.custom.get("compose_path"),
            # "platform": discovery.custom.get("platform"),
        }

    def default_metadata(self, image_name: str | None) -> PackageUpdateInfo:
        relnotes_url: str | None = None
        picture_url: str | None = self.cfg.default_entity_picture_url

        if image_name is not None:
            for pkg in self.common_pkgs.values():
                if pkg.docker is not None and pkg.docker.image_name is not None and pkg.docker.image_name == image_name:
                    self.log.debug(
                        "Found common package", pkg=pkg.docker.image_name, logo_url=picture_url, relnotes_url=relnotes_url
                    )
                    return pkg
            for pkg in self.discovered_pkgs.values():
                if pkg.docker is not None and pkg.docker.image_name is not None and pkg.docker.image_name == image_name:
                    self.log.debug(
                        "Found discovered package", pkg=pkg.docker.image_name, logo_url=picture_url, relnotes_url=relnotes_url
                    )
                    return pkg

        self.log.debug("No common or discovered package found", image_name=image_name)
        return PackageUpdateInfo(
            DockerPackageUpdateInfo(image_name or NO_KNOWN_IMAGE), logo_url=picture_url, release_notes_url=relnotes_url
        )

    def discover_metadata(self) -> dict[str, PackageUpdateInfo]:
        pkgs: dict[str, PackageUpdateInfo] = {}
        cfg = self.cfg.discover_metadata.get("linuxserver.io")
        if cfg and cfg.enabled:
            linuxserver_metadata(pkgs, cache_ttl=cfg.cache_ttl)
        return pkgs


def linuxserver_metadata_api(cache_ttl: int) -> dict:
    """Fetch and cache linuxserver.io API call for image metadata"""
    try:
        with SyncCacheClient(headers=[("cache-control", f"max-age={cache_ttl}")]) as client:
            log.debug(f"Fetching linuxserver.io metadata from API, cache_ttl={cache_ttl}")
            req = client.get("https://api.linuxserver.io/api/v1/images?include_config=false&include_deprecated=false")
            return req.json()
    except Exception:
        log.exception("Failed to fetch linuxserver.io metadata")
        return {}


def linuxserver_metadata(discovered_pkgs: dict[str, PackageUpdateInfo], cache_ttl: int) -> None:
    """Fetch linuxserver.io metadata for all their images via their API"""
    repos: list = linuxserver_metadata_api(cache_ttl).get("data", {}).get("repositories", {}).get("linuxserver", [])
    added = 0
    for repo in repos:
        image_name = repo.get("name")
        if image_name and image_name not in discovered_pkgs:
            discovered_pkgs[image_name] = PackageUpdateInfo(
                DockerPackageUpdateInfo(f"lscr.io/linuxserver/{image_name}"),
                logo_url=repo["project_logo"],
                release_notes_url=f"{repo['github_url']}/releases",
            )
            added += 1
            log.debug("Added linuxserver.io package", pkg=image_name)
    log.info(f"Added {added} linuxserver.io package details")



# python
from collections.abc import AsyncGenerator, Callable
from typing import Any
from unittest.mock import AsyncMock, MagicMock, Mock, patch

import paho.mqtt.client
import pytest
from docker import DockerClient
from docker.models.containers import Container, ContainerCollection
from docker.models.images import Image, RegistryData
from omegaconf import DictConfig, OmegaConf

import updates2mqtt.app
from updates2mqtt.app import (
    App,  # relative import as required
    MqttPublisher,
)
from updates2mqtt.config import Config
from updates2mqtt.model import Discovery, ReleaseProvider


@pytest.fixture
def app_with_mocked_external_dependencies(
    monkeypatch,  # noqa: ANN001
    mock_provider_class: type,
    mock_publisher_class: type,
) -> App:
    cfg: DictConfig = OmegaConf.structured(Config)
    monkeypatch.setattr(updates2mqtt.app, "load_app_config", lambda *_args, **__kwargs: cfg)
    monkeypatch.setattr(updates2mqtt.app, "DockerProvider", mock_provider_class)
    monkeypatch.setattr(updates2mqtt.app, "MqttPublisher", mock_publisher_class)
    app: App = App()
    return app


@pytest.fixture
def mock_discoveries(mock_provider: ReleaseProvider) -> list[Discovery]:
    return [Discovery(mock_provider, "thing-1", "test001")]


@pytest.fixture
def mock_discovery_generator(mock_discoveries: list[Discovery]) -> Callable[..., AsyncGenerator[Discovery, Any]]:
    async def g(*args: Any) -> AsyncGenerator[Discovery]:  # noqa: ARG001
        for d in mock_discoveries:
            yield d

    return g


@pytest.fixture
def mock_provider_class(mock_provider: ReleaseProvider) -> type:
    class MockReleaseProvider(ReleaseProvider):
        def __new__(cls, *args: Any, **kwargs: Any) -> ReleaseProvider:  # type: ignore[misc] # noqa: ARG004
            return mock_provider

    return MockReleaseProvider


@pytest.fixture
def mock_publisher_class(mock_publisher: MqttPublisher) -> type:
    class MockPublisher(MqttPublisher):
        def __new__(cls, *args: Any, **kwargs: Any) -> MqttPublisher:  # type: ignore[misc] # noqa: ARG004
            return mock_publisher

    return MockPublisher


@pytest.fixture
def mock_provider() -> ReleaseProvider:
    provider: ReleaseProvider = AsyncMock(spec=ReleaseProvider)
    provider.source_type = "unit_test"
    provider.command.return_value = True  # type: ignore[attr-defined]
    provider.resolve.return_value = Discovery(  # type: ignore[attr-defined]
        provider, "fooey", session="test-mqtt-123", current_version="v2", latest_version="v2"
    )
    provider.hass_state_format.return_value = {"fixture": "test_exec"}  # type: ignore[attr-defined]
    return provider


@pytest.fixture
def mock_publisher(mock_mqtt_client: paho.mqtt.client.Client) -> MqttPublisher:
    publisher: MqttPublisher = AsyncMock(MqttPublisher)
    publisher.client = mock_mqtt_client
    return publisher


@pytest.fixture
def mock_mqtt_client() -> paho.mqtt.client.Client:
    return MagicMock(spec=paho.mqtt.client.Client, name="MQTT Client Fixture")


@pytest.fixture
def mock_docker_client() -> DockerClient:
    client = Mock(spec=DockerClient)
    coll = Mock(spec=ContainerCollection)

    def reg_data_select(v: str) -> RegistryData:
        reg_data = Mock(spec=RegistryData)
        match v:
            case "testy/mctest:latest":
                reg_data.short_id = "sha256:c5385387575"
            case "testy/mctest":
                reg_data.short_id = "sha256:9e2bbca07938"
            case "ubuntu":
                reg_data.short_id = "sha256:85a5385853bd"
            case _:
                reg_data.short_id = "sha256:999999999999"
        return reg_data

    client.images.get_registry_data = Mock(side_effect=reg_data_select)

    client.containers = coll
    coll.list.return_value = [
        build_mock_container("testy/mctest:latest", opsys="macos"),
        build_mock_container("ubuntu"),
        build_mock_container("common/pkg"),
        build_mock_container(
            "testy/mctest",
            picture="https://piccy",
            relnotes="https://release",
            arch="amd64",
        ),
    ]
    patch("docker.from_env", return_value=client)
    return client


def build_mock_container(
    tag: str, picture: str | None = None, relnotes: str | None = None, opsys: str = "linux", arch: str = "arm64"
) -> Container:
    c = Mock(spec=Container)
    c.image = Mock(spec=Image)
    c.image.tags = [tag]
    c.image.attrs = {}
    c.image.attrs["Os"] = opsys
    c.image.attrs["Architecture"] = arch
    bare_tag = tag.split(":")[0]
    long_hash = "9e2bbca079387d7965c3a9cee6d0c53f4f4e63ff7637877a83c4c05f2a666112"
    c.image.attrs["RepoDigests"] = [f"{bare_tag}@sha256:{long_hash}"]
    c.attrs = {}
    c.attrs["Config"] = {}
    c.attrs["Config"]["Env"] = []
    if picture:
        c.attrs["Config"]["Env"].append(f"UPD2MQTT_PICTURE={picture}")
    if relnotes:
        c.attrs["Config"]["Env"].append(f"UPD2MQTT_RELNOTES={relnotes}")
    return c



import datetime
from pathlib import Path

from pytest_subprocess import FakeProcess  # type: ignore[import-not-found]

from updates2mqtt.integrations.git_utils import git_check_update_available, git_pull, git_timestamp, git_trust

GIT_EXEC = Path("/usr/bin/git")


def test_git_timestamp(fake_process: FakeProcess) -> None:
    fake_process.register([fake_process.any()], stdout="""2024-04-12T00:16:33+01:00""")
    assert git_timestamp(Path("/my/path"), GIT_EXEC) == datetime.datetime(
        2024,
        4,
        12,
        0,
        16,
        33,
        0,
        datetime.timezone(offset=datetime.timedelta(hours=1)),
    )


def test_git_trust(fake_process: FakeProcess) -> None:
    fake_process.register("/usr/bin/git config --global --add safe.directory /my/path", returncode=0)
    assert git_trust(Path("/my/path"), GIT_EXEC)


def test_git_pull(fake_process: FakeProcess) -> None:
    fake_process.register("/usr/bin/git pull", returncode=0)
    assert git_pull(Path("/my/path"), GIT_EXEC)
    fake_process.register("/usr/bin/git pull", returncode=23)
    assert git_pull(Path("/my/path"), GIT_EXEC) is False


def test_git_check_update_available(fake_process: FakeProcess) -> None:
    fake_process.register(
        "/usr/bin/git fetch;/usr/bin/git status -uno",
        stdout="Your branch is behind 'origin/main' by 1 commit, and can be fast-forwarded.",
        returncode=0,
    )
    assert git_check_update_available(Path("/my/path"), GIT_EXEC)
    fake_process.register(
        "/usr/bin/git fetch;/usr/bin/git status -uno", stdout="Your branch is up to date with 'origin/main'.", returncode=0
    )
    assert git_check_update_available(Path("/my/path"), GIT_EXEC) is False
    fake_process.register("/usr/bin/git fetch;/usr/bin/git status -uno", returncode=1)
    assert git_check_update_available(Path("/my/path"), GIT_EXEC, timeout=5) is False



from abc import abstractmethod
from collections.abc import AsyncGenerator, Callable
from threading import Event
from typing import Any

import structlog


class Discovery:
    """Discovered component from a scan"""

    def __init__(
        self,
        provider: "ReleaseProvider",
        name: str,
        session: str,
        entity_picture_url: str | None = None,
        current_version: str | None = None,
        latest_version: str | None = None,
        can_update: bool = False,
        can_build: bool = False,
        can_restart: bool = False,
        status: str = "on",
        update_policy: str | None = None,
        update_last_attempt: float | None = None,
        release_url: str | None = None,
        release_summary: str | None = None,
        title_template: str = "Update for {name} on {node}",
        device_icon: str | None = None,
        custom: dict[str, Any] | None = None,
        features: list[str] | None = None,
    ) -> None:
        self.provider: ReleaseProvider = provider
        self.source_type: str = provider.source_type
        self.session: str = session
        self.name: str = name
        self.entity_picture_url: str | None = entity_picture_url
        self.current_version: str | None = current_version
        self.latest_version: str | None = latest_version
        self.can_update: bool = can_update
        self.can_build: bool = can_build
        self.can_restart: bool = can_restart
        self.release_url: str | None = release_url
        self.release_summary: str | None = release_summary
        self.title_template: str | None = title_template
        self.device_icon: str | None = device_icon
        self.status: str = status
        self.update_policy: str | None = update_policy
        self.update_last_attempt: float | None = update_last_attempt
        self.custom: dict[str, Any] = custom or {}
        self.features: list[str] = features or []

    def __repr__(self) -> str:
        """Build a custom string representation"""
        return f"Discovery('{self.name}','{self.source_type}',current={self.current_version},latest={self.latest_version})"


class ReleaseProvider:
    """Abstract base class for release providers, such as container scanners or package managers API calls"""

    def __init__(self, source_type: str = "base") -> None:
        self.source_type: str = source_type
        self.discoveries: dict[str, Discovery] = {}
        self.log: Any = structlog.get_logger().bind(integration=self.source_type)
        self.stopped = Event()

    def stop(self) -> None:
        """Stop any loops or background tasks"""
        self.log.info("Asking release provider to stop", source_type=self.source_type)
        self.stopped.set()

    @abstractmethod
    def update(self, discovery: Discovery) -> bool:
        """Attempt to update the component version"""

    @abstractmethod
    def rescan(self, discovery: Discovery) -> Discovery | None:
        """Rescan a previously discovered component"""

    @abstractmethod
    async def scan(self, session: str) -> AsyncGenerator[Discovery]:
        """Scan for components to monitor"""
        raise NotImplementedError
        # force recognition as an async generator
        if False:  # type: ignore[unreachable]
            yield 0

    def hass_config_format(self, discovery: Discovery) -> dict:
        _ = discovery
        return {}

    def hass_state_format(self, discovery: Discovery) -> dict:
        _ = discovery
        return {}

    @abstractmethod
    def command(self, discovery_name: str, command: str, on_update_start: Callable, on_update_end: Callable) -> bool:
        """Execute a command on a discovered component"""

    @abstractmethod
    def resolve(self, discovery_name: str) -> Discovery | None:
        """Resolve a discovered component by name"""



from pathlib import Path
from unittest.mock import patch

import pytest
from docker import DockerClient
from pytest_httpx import HTTPXMock
from pytest_subprocess import FakeProcess  # type: ignore[import-not-found]

import updates2mqtt.integrations.docker as mut
from updates2mqtt.config import DockerPackageUpdateInfo, MetadataSourceConfig
from updates2mqtt.model import Discovery


async def test_scanner(mock_docker_client: DockerClient) -> None:
    with patch("docker.from_env", return_value=mock_docker_client):
        uut = mut.DockerProvider(mut.DockerConfig(discover_metadata={}), mut.UpdateInfoConfig(), mut.NodeConfig())
        session = "unit_123"
        results: list[Discovery] = [d async for d in uut.scan(session)]

    unchanged: list[Discovery] = [d for d in results if d.current_version == d.latest_version]
    assert len(unchanged) == 1
    assert unchanged[0].entity_picture_url == "https://piccy"
    assert unchanged[0].release_url == "https://release"
    assert unchanged[0].custom["platform"] == "linux/amd64"
    changed = [d for d in results if d.current_version != d.latest_version]
    assert len(changed) == 3


async def test_common_packages(mock_docker_client: DockerClient) -> None:
    with patch("docker.from_env", return_value=mock_docker_client):
        uut = mut.DockerProvider(mut.DockerConfig(discover_metadata={}), mut.UpdateInfoConfig(), mut.NodeConfig())
        uut.common_pkgs = {
            "common_pkg": mut.PackageUpdateInfo(
                docker=DockerPackageUpdateInfo(image_name="common/pkg"),
                logo_url="https://commonhub/pkg/logo",
                release_notes_url="https://commonhub/pkg/logo",
            )
        }
        session = "unit_123"
        results: list[Discovery] = [d async for d in uut.scan(session)]

    common: list[Discovery] = [d for d in results if d.custom["image_ref"] == "common/pkg"]
    assert len(common) == 1
    assert common[0].entity_picture_url == "https://commonhub/pkg/logo"
    assert common[0].release_url == "https://commonhub/pkg/logo"


@pytest.mark.httpx_mock(assert_all_requests_were_expected=False)
def test_discover_metadata(httpx_mock: HTTPXMock, mock_docker_client: DockerClient) -> None:
    httpx_mock.add_response(
        json={
            "data": {
                "repositories": {
                    "linuxserver": [
                        {
                            "name": "mctesty901",
                            "project_logo": "http://logos/mctesty.png",
                            "github_url": "https://github/mctesty/901",
                        }
                    ]
                }
            }
        }
    )
    with patch("docker.from_env", return_value=mock_docker_client):
        uut = mut.DockerProvider(
            mut.DockerConfig(discover_metadata={"linuxserver.io": MetadataSourceConfig(enabled=True, cache_ttl=0)}),
            mut.UpdateInfoConfig(),
            mut.NodeConfig(),
        )
        uut.discover_metadata()
    assert "mctesty901" in uut.discovered_pkgs
    pkg = uut.discovered_pkgs["mctesty901"]
    assert pkg.docker is not None
    assert pkg.docker.image_name == "lscr.io/linuxserver/mctesty901"
    assert pkg.logo_url == "http://logos/mctesty.png"
    assert pkg.release_notes_url == "https://github/mctesty/901/releases"


def test_build(mock_docker_client: DockerClient, fake_process: FakeProcess, tmpdir: Path) -> None:
    with patch("docker.from_env", return_value=mock_docker_client):
        uut = mut.DockerProvider(mut.DockerConfig(discover_metadata={}), mut.UpdateInfoConfig(), mut.NodeConfig())
        d = Discovery(uut, "build-test-dummy", session="test-123")
        fake_process.register("docker compose build", returncode=0)
        assert uut.build(d, str(tmpdir))
        fake_process.register("docker compose build", returncode=33)
        assert not uut.build(d, str(tmpdir))



import asyncio
import json
import time
from unittest import mock
from unittest.mock import Mock, patch

import paho.mqtt.client
import pytest
from paho.mqtt.client import MQTTMessage

from updates2mqtt.config import HomeAssistantConfig, MqttConfig, NodeConfig
from updates2mqtt.model import Discovery, ReleaseProvider
from updates2mqtt.mqtt import MqttPublisher


@pytest.mark.parametrize("protocol", ["3", "3.1", "5", "?"])
def test_publish(mock_mqtt_client: Mock, protocol: str) -> None:
    config = MqttConfig(protocol=protocol)
    hass_config = HomeAssistantConfig()
    node_config = NodeConfig()

    with patch.object(paho.mqtt.client.Client, "__new__", lambda *_args, **_kwargs: mock_mqtt_client):
        uut = MqttPublisher(config, node_config, hass_config)
        uut.start()

        uut.publish("test.topic.123", {"foo": "a8", "bar": False})
        mock_mqtt_client.connect.assert_called_once()
        mock_mqtt_client.publish.assert_called_with("test.topic.123", payload='{"foo": "a8", "bar": false}', qos=0, retain=True)


@pytest.mark.asyncio
async def test_handler(mock_mqtt_client: Mock) -> None:
    config = MqttConfig()
    hass_config = HomeAssistantConfig()
    node_config = NodeConfig()
    node_config.name = "testing"
    with patch("updates2mqtt.mqtt.mqtt.Client", new=mock_mqtt_client):
        uut = MqttPublisher(config, node_config, hass_config)
        uut.start(event_loop=asyncio.get_running_loop())

        provider = Mock(spec=ReleaseProvider)
        provider.source_type = "unit_test"
        discovery = Discovery(provider, "qux", session="test-mqtt-123")
        provider.command.return_value = discovery
        provider.hass_state_format.return_value = {}

        topic_name = uut.subscribe_hass_command(provider)
        mock_message = Mock()
        mock_message.topic = topic_name
        mock_message.payload = "|".join([provider.source_type, "qux", "install"])
        uut.handle_message(mock_message)

        cutoff = time.time() + 10
        while time.time() <= cutoff and not provider.command.called:  # noqa: ASYNC110
            await asyncio.sleep(0.5)

        provider.command.assert_called_with("qux", "install", mock.ANY, mock.ANY)


async def test_execute_command_remote(mock_mqtt_client: Mock, mock_provider: ReleaseProvider) -> None:
    config = MqttConfig()
    hass_config = HomeAssistantConfig()
    node_config = NodeConfig("TESTBED")

    with patch.object(paho.mqtt.client.Client, "__new__", lambda *_args, **_kwargs: mock_mqtt_client):
        uut = MqttPublisher(config, node_config, hass_config)
        uut.start(event_loop=asyncio.get_running_loop())

        uut.subscribe_hass_command(mock_provider)
        dummy_callable = lambda: None  # noqa: E731

        mqtt_bytes_msg = MQTTMessage(topic=b"updates2mqtt/TESTBED/unit_test")
        mqtt_bytes_msg.payload = b"unit_test|fooey|install"
        await uut.execute_command(mqtt_bytes_msg, dummy_callable, dummy_callable)

        mock_mqtt_client.publish.assert_called_with(
            "updates2mqtt/TESTBED/unit_test/fooey",
            payload=json.dumps(
                {
                    "installed_version": "v2",
                    "latest_version": "v2",
                    "title": "Update for fooey on TESTBED",
                    "in_progress": True,
                }
            ),
            qos=0,
            retain=True,
        )


@pytest.mark.asyncio
async def test_execute_command_local(mock_mqtt_client: Mock, mock_provider: ReleaseProvider) -> None:
    config = MqttConfig()
    hass_config = HomeAssistantConfig()
    node_config = NodeConfig("TESTBED")

    with patch.object(paho.mqtt.client.Client, "__new__", lambda *_args, **_kwargs: mock_mqtt_client):
        uut = MqttPublisher(config, node_config, hass_config)

        uut.start(event_loop=asyncio.get_running_loop())

        uut.subscribe_hass_command(mock_provider)

        discovery = Discovery(mock_provider, "fooey", session="test-mqtt-123", current_version="v1")
        uut.local_message(discovery, "install")
        await asyncio.sleep(1)

        mock_mqtt_client.publish.assert_called_with(
            "updates2mqtt/TESTBED/unit_test/fooey",
            payload=json.dumps(
                {
                    "installed_version": "v2",
                    "latest_version": "v2",
                    "title": "Update for fooey on TESTBED",
                    "in_progress": True,
                }
            ),
            qos=0,
            retain=True,
        )


@pytest.mark.asyncio
async def test_stop(mock_mqtt_client: Mock, mock_provider: ReleaseProvider) -> None:
    config = MqttConfig()
    hass_config = HomeAssistantConfig()
    node_config = NodeConfig()

    with patch.object(paho.mqtt.client.Client, "__new__", lambda *_args, **_kwargs: mock_mqtt_client):
        uut = MqttPublisher(config, node_config, hass_config)

        uut.start(event_loop=asyncio.get_running_loop())

        uut.subscribe_hass_command(mock_provider)

        uut.stop()
        await asyncio.sleep(1)
        mock_mqtt_client.loop_stop.assert_called()
        mock_mqtt_client.disconnect.assert_called()



import asyncio
import json
import time
from collections.abc import Callable
from dataclasses import dataclass, field
from threading import Event
from typing import Any

import paho.mqtt.client as mqtt
import paho.mqtt.subscribeoptions
import structlog
from paho.mqtt.client import MQTT_CLEAN_START_FIRST_ONLY, MQTTMessage
from paho.mqtt.enums import CallbackAPIVersion, MQTTErrorCode, MQTTProtocolVersion
from paho.mqtt.properties import Properties
from paho.mqtt.reasoncodes import ReasonCode

from updates2mqtt.model import Discovery, ReleaseProvider

from .config import HomeAssistantConfig, MqttConfig, NodeConfig
from .hass_formatter import hass_format_config, hass_format_state

log = structlog.get_logger()


@dataclass
class LocalMessage:
    topic: str | None = field(default=None)
    payload: str | None = field(default=None)


class MqttPublisher:
    def __init__(self, cfg: MqttConfig, node_cfg: NodeConfig, hass_cfg: HomeAssistantConfig) -> None:
        self.cfg: MqttConfig = cfg
        self.node_cfg: NodeConfig = node_cfg
        self.hass_cfg: HomeAssistantConfig = hass_cfg
        self.providers_by_topic: dict[str, ReleaseProvider] = {}
        self.event_loop: asyncio.AbstractEventLoop | None = None
        self.client: mqtt.Client | None = None
        self.fatal_failure = Event()
        self.log = structlog.get_logger().bind(host=cfg.host, integration="mqtt")

    def start(self, event_loop: asyncio.AbstractEventLoop | None = None) -> None:
        logger = self.log.bind(action="start")
        try:
            protocol: MQTTProtocolVersion
            if self.cfg.protocol in ("3", "3.11"):
                protocol = MQTTProtocolVersion.MQTTv311
            elif self.cfg.protocol == "3.1":
                protocol = MQTTProtocolVersion.MQTTv31
            elif self.cfg.protocol in ("5", "5.0"):
                protocol = MQTTProtocolVersion.MQTTv5
            else:
                self.log.info("No valid MQTT protocol version found (%s), setting to default v3.11", self.cfg.protocol)
                protocol = MQTTProtocolVersion.MQTTv311
            self.log.debug("MQTT protocol set to %r", protocol)

            self.event_loop = event_loop or asyncio.get_event_loop()
            self.client = mqtt.Client(
                callback_api_version=CallbackAPIVersion.VERSION2,
                client_id=f"updates2mqtt_{self.node_cfg.name}",
                clean_session=True if protocol != MQTTProtocolVersion.MQTTv5 else None,
                protocol=protocol,
            )
            self.client.username_pw_set(self.cfg.user, password=self.cfg.password)
            rc: MQTTErrorCode = self.client.connect(
                host=self.cfg.host,
                port=self.cfg.port,
                keepalive=60,
                clean_start=MQTT_CLEAN_START_FIRST_ONLY,
            )
            self.log.info("Client connection requested", result_code=rc)

            self.client.on_connect = self.on_connect
            self.client.on_disconnect = self.on_disconnect
            self.client.on_message = self.on_message
            self.client.on_subscribe = self.on_subscribe
            self.client.on_unsubscribe = self.on_unsubscribe

            self.client.loop_start()

            logger.info("Connected to broker", host=self.cfg.host, port=self.cfg.port)
        except Exception as e:
            logger.error("Failed to connect to broker", host=self.cfg.host, port=self.cfg.port, error=str(e))
            raise OSError(f"Connection Failure to {self.cfg.host}:{self.cfg.port} as {self.cfg.user} -- {e}") from e

    def stop(self) -> None:
        if self.client:
            self.client.loop_stop()
            self.client.disconnect()
            self.client = None

    def is_available(self) -> bool:
        return self.client is not None and not self.fatal_failure.is_set()

    def on_connect(
        self, _client: mqtt.Client, _userdata: Any, _flags: mqtt.ConnectFlags, rc: ReasonCode, _props: Properties | None
    ) -> None:
        if not self.client or self.fatal_failure.is_set():
            self.log.warn("No client, check if started and authorized")
            return
        if rc.getName() == "Not authorized":
            self.fatal_failure.set()
            log.error("Invalid MQTT credentials", result_code=rc)

        self.log.info("Connected to broker", result_code=rc)
        for topic, provider in self.providers_by_topic.items():
            self.log.info("(Re)subscribing", topic=topic, provider=provider.source_type)
            self.client.subscribe(topic)

    def on_disconnect(
        self,
        _client: mqtt.Client,
        _userdata: Any,
        _disconnect_flags: mqtt.DisconnectFlags,
        rc: ReasonCode,
        _props: Properties | None,
    ) -> None:
        self.log.info("Disconnected from broker", result_code=rc)

    async def clean_topics(
        self, provider: ReleaseProvider, last_scan_session: str | None, wait_time: int = 5, force: bool = False
    ) -> None:
        logger = self.log.bind(action="clean")
        if self.fatal_failure.is_set():
            return
        logger.info("Starting clean cycle")
        cleaner = mqtt.Client(
            callback_api_version=CallbackAPIVersion.VERSION1,
            client_id=f"updates2mqtt_clean_{self.node_cfg.name}",
            clean_session=True,
        )
        results = {"cleaned": 0, "handled": 0, "discovered": 0, "last_timestamp": time.time()}
        cleaner.username_pw_set(self.cfg.user, password=self.cfg.password)
        cleaner.connect(host=self.cfg.host, port=self.cfg.port, keepalive=60)
        prefixes = [
            f"{self.hass_cfg.discovery.prefix}/update/{self.node_cfg.name}_{provider.source_type}_",
            f"{self.cfg.topic_root}/{self.node_cfg.name}/{provider.source_type}/",
        ]

        def cleanup(_client: mqtt.Client, _userdata: Any, msg: mqtt.MQTTMessage) -> None:
            if msg.retain and any(msg.topic.startswith(prefix) for prefix in prefixes):
                session = None
                results["discovered"] += 1
                try:
                    payload = self.safe_json_decode(msg.payload)
                    session = payload.get("source_session")
                except Exception as e:
                    log.warn(
                        "Unable to handle payload for %s: %s",
                        msg.topic,
                        e,
                        exc_info=1,
                    )
                results["handled"] += 1
                results["last_timestamp"] = time.time()
                if session is not None and last_scan_session is not None and session != last_scan_session:
                    log.debug("Removing stale msg", topic=msg.topic, session=session)
                    cleaner.publish(msg.topic, "", retain=True)
                    results["cleaned"] += 1
                elif session is None and force:
                    log.debug("Removing untrackable msg", topic=msg.topic)
                    cleaner.publish(msg.topic, "", retain=True)
                    results["cleaned"] += 1
                else:
                    log.debug(
                        "Retaining topic with current session: %s",
                        msg.topic,
                    )
            else:
                log.debug("Skipping clean of %s", msg.topic)

        cleaner.on_message = cleanup
        options = paho.mqtt.subscribeoptions.SubscribeOptions(noLocal=True)
        cleaner.subscribe(f"{self.hass_cfg.discovery.prefix}/update/#", options=options)
        cleaner.subscribe(f"{self.cfg.topic_root}/{self.node_cfg.name}/{provider.source_type}/#", options=options)

        while time.time() - results["last_timestamp"] <= wait_time:
            cleaner.loop(0.5)

        log.info(
            f"Clean completed, discovered:{results['discovered']}, handled:{results['handled']}, cleaned:{results['cleaned']}"
        )

    def safe_json_decode(self, jsonish: str | bytes | None) -> dict:
        if jsonish is None:
            return {}
        try:
            return json.loads(jsonish)
        except Exception:
            log.exception("JSON decode fail (%s); %s", jsonish)
        try:
            return json.loads(jsonish[1:-1])
        except Exception:
            log.exception("JSON decode fail (%s): %s", jsonish[1:-1])
        return {}

    async def execute_command(
        self, msg: MQTTMessage | LocalMessage, on_update_start: Callable, on_update_end: Callable
    ) -> None:
        logger = self.log.bind(topic=msg.topic, payload=msg.payload)
        comp_name: str | None = None
        command: str | None = None
        try:
            logger.info("Execution starting")
            source_type: str | None = None

            payload: str | None = None
            if isinstance(msg.payload, bytes):
                payload = msg.payload.decode("utf-8")
            elif isinstance(msg.payload, str):
                payload = msg.payload
            if payload and "|" in payload:
                source_type, comp_name, command = payload.split("|")

            provider: ReleaseProvider | None = self.providers_by_topic.get(msg.topic) if msg.topic else None
            if not provider:
                logger.warn("Unexpected provider type %s", msg.topic)
            elif provider.source_type != source_type:
                logger.warn("Unexpected source type %s", source_type)
            elif command != "install" or not comp_name:
                logger.warn("Invalid payload in command message: %s", msg.payload)
            else:
                logger.info(
                    "Passing %s command to %s scanner for %s",
                    command,
                    source_type,
                    comp_name,
                )
                updated = provider.command(comp_name, command, on_update_start, on_update_end)
                discovery = provider.resolve(comp_name)
                if updated and discovery:
                    self.publish_hass_state(discovery, updated)
                else:
                    logger.debug("No change to republish after execution")
            logger.info("Execution ended")
        except Exception:
            logger.exception("Execution failed", component=comp_name, command=command)

    def local_message(self, discovery: Discovery, command: str) -> None:
        """Simulate an incoming MQTT message for local commands"""
        msg = LocalMessage(
            topic=self.command_topic(discovery.provider), payload="|".join([discovery.source_type, discovery.name, command])
        )
        self.handle_message(msg)

    def on_subscribe(
        self,
        _client: mqtt.Client,
        userdata: Any,
        mid: int,
        reason_code_list: list[ReasonCode],
        properties: Properties | None = None,
    ) -> None:
        self.log.debug(
            "on_subscribe, userdata=%s, mid=%s, reasons=%s, properties=%s", userdata, mid, reason_code_list, properties
        )

    def on_unsubscribe(
        self,
        _client: mqtt.Client,
        userdata: Any,
        mid: int,
        reason_code_list: list[ReasonCode],
        properties: Properties | None = None,
    ) -> None:
        self.log.debug(
            "on_unsubscribe, userdata=%s, mid=%s, reasons=%s, properties=%s", userdata, mid, reason_code_list, properties
        )

    def on_message(self, _client: mqtt.Client, _userdata: Any, msg: mqtt.MQTTMessage) -> None:
        """Callback for incoming MQTT messages"""  # noqa: D401
        if msg.topic in self.providers_by_topic:
            self.handle_message(msg)
        else:
            # apparently the root non-wildcard sub sometimes brings in child topics
            self.log.debug("Unhandled message #%s on %s:%s", msg.mid, msg.topic, msg.payload)

    def handle_message(self, msg: mqtt.MQTTMessage | LocalMessage) -> None:
        def update_start(discovery: Discovery) -> None:
            self.publish_hass_state(discovery, in_progress=True)

        def update_end(discovery: Discovery) -> None:
            self.publish_hass_state(discovery, in_progress=False)

        if self.event_loop is not None:
            asyncio.run_coroutine_threadsafe(self.execute_command(msg, update_start, update_end), self.event_loop)
        else:
            self.log.error("No event loop to handle message", topic=msg.topic)

    def config_topic(self, discovery: Discovery) -> str:
        prefix = self.hass_cfg.discovery.prefix
        return f"{prefix}/update/{self.node_cfg.name}_{discovery.source_type}_{discovery.name}/update/config"

    def state_topic(self, discovery: Discovery) -> str:
        return f"{self.cfg.topic_root}/{self.node_cfg.name}/{discovery.source_type}/{discovery.name}"

    def command_topic(self, provider: ReleaseProvider) -> str:
        return f"{self.cfg.topic_root}/{self.node_cfg.name}/{provider.source_type}"

    def publish_hass_state(self, discovery: Discovery, in_progress: bool = False) -> None:
        self.log.debug("HASS State update, in progress: %s, discovery: %s", in_progress, discovery)
        self.publish(
            self.state_topic(discovery),
            hass_format_state(
                discovery,
                self.node_cfg.name,
                discovery.session,
                in_progress=in_progress,
            ),
        )

    def publish_hass_config(self, discovery: Discovery) -> None:
        object_id = f"{discovery.source_type}_{self.node_cfg.name}_{discovery.name}"
        command_topic: str | None = self.command_topic(discovery.provider) if discovery.can_update else None
        self.publish(
            self.config_topic(discovery),
            hass_format_config(
                discovery,
                object_id,
                self.node_cfg.name,
                self.state_topic(discovery),
                command_topic,
                discovery.session,
            ),
        )

    def subscribe_hass_command(self, provider: ReleaseProvider):  # noqa: ANN201
        topic = self.command_topic(provider)
        if topic in self.providers_by_topic or self.client is None:
            self.log.debug("Skipping subscription", topic=topic)
        else:
            self.log.info("Handler subscribing", topic=topic)
            self.providers_by_topic[topic] = provider
            self.client.subscribe(topic)
        return topic

    def loop_once(self) -> None:
        if self.client:
            self.client.loop()

    def publish(self, topic: str, payload: dict, qos: int = 0, retain: bool = True) -> None:
        if self.client:
            self.client.publish(topic, payload=json.dumps(payload), qos=qos, retain=retain)



# python
import asyncio
import signal
import time
import types
from collections.abc import AsyncGenerator, Coroutine
from typing import Any, NoReturn
from unittest.mock import ANY, call

import pytest

from updates2mqtt.app import App, run  # relative import as required
from updates2mqtt.model import Discovery


async def test_scan(
    app_with_mocked_external_dependencies: App,
    mock_discoveries: list[Discovery],
    mock_discovery_generator: AsyncGenerator[Discovery],
    monkeypatch,  # noqa: ANN001
) -> None:
    uut: App = app_with_mocked_external_dependencies
    monkeypatch.setattr(uut.scanners[0], "scan", mock_discovery_generator)
    await uut.scan()
    uut.publisher.clean_topics.assert_has_calls(  # type: ignore[attr-defined]
        [call(uut.scanners[0], None, force=True), call(uut.scanners[0], ANY, force=False)]
    )
    uut.publisher.publish_hass_state.assert_has_calls([call(d) for d in mock_discoveries])  # type: ignore[attr-defined]


async def test_main_loop(
    app_with_mocked_external_dependencies: App,
    mock_discovery_generator: AsyncGenerator[Discovery],
    monkeypatch,  # noqa: ANN001
) -> None:
    uut: App = app_with_mocked_external_dependencies
    monkeypatch.setattr(uut.scanners[0], "scan", mock_discovery_generator)
    start_time = time.time()
    monkeypatch.setattr(uut.publisher, "is_available", lambda: time.time() > start_time + 3)

    with pytest.raises(SystemExit):
        await uut.main_loop()
    uut.publisher.assert_has_calls(  # type: ignore[attr-defined]
        [
            call.start(),
            call.subscribe_hass_command(uut.scanners[0]),
            call.stop(),
        ]
    )  # pyright: ignore[reportAttributeAccessIssue]


class DummyApp:
    """Dummy App to replace updates2mqtt.app.App during tests.

    Records the created instance on DummyApp.instance for assertions.
    """

    instance = None

    def __init__(self) -> None:
        DummyApp.instance = self
        self.run_called = False
        self.shutdown_called = False

    async def main_loop(self) -> None:
        # an async method to mirror the real App.run signature
        self.run_called = True

    def shutdown(self, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        self.shutdown_called = True


def test_run_sets_signal_and_calls_asyncio_run(monkeypatch) -> None:  # noqa: ANN001
    calls: dict[str, Any] = {}

    # Replace signal.signal so we capture its arguments
    def fake_signal(sig: int, handler: types.MethodType) -> None:
        calls["sig"] = sig
        calls["handler"] = handler

    monkeypatch.setattr(signal, "signal", fake_signal)

    # Patch the App class in the updates2mqtt.app module to DummyApp
    import updates2mqtt.app as app_module

    monkeypatch.setattr(app_module, "App", DummyApp)

    # Patch asyncio.run to record that it was called with a coroutine and debug flag
    def fake_asyncio_run(coro: Coroutine, debug: bool = False) -> None:
        calls["coro"] = coro
        calls["debug"] = debug
        return

    monkeypatch.setattr(asyncio, "run", fake_asyncio_run)

    # Execute the run function under test
    run()

    # Assertions:
    # - signal.signal was called with SIGTERM and the DummyApp.instance.shutdown bound method
    assert calls.get("sig") == signal.SIGTERM
    assert DummyApp.instance is not None
    assert calls.get("handler") == DummyApp.instance.shutdown
    # debug flag forwarded as False in run()
    assert calls.get("debug") is False
    # - asyncio.run was called with a coroutine object (the result of DummyApp.instance.run())
    coro = calls.get("coro")
    assert isinstance(coro, types.CoroutineType)


def test_run_handles_asyncio_cancellederror(monkeypatch) -> None:  # noqa: ANN001
    # Ensure App is replaced so run() will create DummyApp without side effects
    import updates2mqtt.app as app_module

    monkeypatch.setattr(app_module, "App", DummyApp)

    # Patch signal.signal to a noop to avoid altering test process handlers
    monkeypatch.setattr(signal, "signal", lambda *args, **kwargs: None)  # noqa: ARG005

    # Patch asyncio.run to raise CancelledError to exercise the except branch
    def raising_asyncio_run(_coro: Coroutine, debug: bool = False) -> NoReturn:  # noqa: ARG001
        raise asyncio.CancelledError()

    monkeypatch.setattr(asyncio, "run", raising_asyncio_run)

    # Call run(); should not raise CancelledError (handled inside run)
    run()

    # If we reached here, the CancelledError was handled; also ensure DummyApp was instantiated
    assert DummyApp.instance is not None



import datetime
import subprocess
from pathlib import Path

import structlog

log = structlog.get_logger()


def git_trust(repo_path: Path, git_path: Path) -> bool:
    try:
        subprocess.run(f"{git_path} config --global --add safe.directory {repo_path}", check=True, shell=True, cwd=repo_path)
        return True
    except Exception as e:
        log.warn("GIT Unable to trust repo at %s: %s", repo_path, e)
        return False


def git_timestamp(repo_path: Path, git_path: Path) -> datetime.datetime | None:
    result = None
    try:
        result = subprocess.run(
            str(git_path) + r" log -1 --format=%cI --no-show-signature",
            cwd=repo_path,
            shell=True,
            text=True,
            capture_output=True,
            check=True,
        )
        return datetime.datetime.fromisoformat(result.stdout.strip())
    except subprocess.CalledProcessError as cpe:
        log.warn("GIT No result from git log at %s: %s", repo_path, cpe)
    except Exception as e:
        log.error("GIT Unable to parse timestamp at %s - %s: %s", repo_path, result.stdout if result else "<NO RESULT>", e)
    return None


def git_check_update_available(repo_path: Path, git_path: Path, timeout: int = 120) -> bool:
    result = None
    try:
        # check if remote repo ahead
        result = subprocess.run(
            f"{git_path} fetch;{git_path} status -uno",
            capture_output=True,
            text=True,
            shell=True,
            check=True,
            cwd=repo_path,
            timeout=timeout,
        )
        if result.returncode == 0 and "Your branch is behind" in result.stdout:
            log.info("Local git repo update available", path=repo_path, status=result.stdout.strip())
            return True
    except Exception as e:
        log.warn("GIT Unable to check status %s: %s", result.stdout if result else "<NO RESULT>", e)
    return False


def git_pull(repo_path: Path, git_path: Path) -> bool:
    log.info("GIT Pulling git at %s", repo_path)
    proc = subprocess.run(f"{git_path} pull", shell=True, check=False, cwd=repo_path, timeout=300)
    if proc.returncode == 0:
        log.info("GIT pull at %s successful", repo_path)
        return True
    log.warn("GIT pull at %s failed: %s", repo_path, proc.returncode)
    return False
